#!/usr/bin/env python
"""
desc goes here
"""
import json
import logging
import os
import requests
import time
import uuid


from requests import HTTPError
from urllib.parse import urljoin, quote


class IngestApi:
    def __init__(self, url=None, ingest_api_root=None):
        format = '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        logging.getLogger("requests").setLevel(logging.WARNING)
        self.logger = logging.getLogger(__name__)

        if not url and 'INGEST_API' in os.environ:
            url = os.environ['INGEST_API']
            # expand interpolated env vars
            url = os.path.expandvars(url)
            self.logger.info("using " + url + " for ingest API")
        self.url = url if url else "http://localhost:8080"

        self.headers = {'Content-type': 'application/json'}
        self.submission_links = {}
        self.token = None
        self.ingest_api_root = ingest_api_root if ingest_api_root is not None else self.get_root_url()

    def set_token(self, token):
        self.token = token

    def get_root_url(self):
        reply = requests.get(self.url, headers=self.headers)
        return reply.json()["_links"]

    def _get_url_for_link(self, url, link_name):
        r = requests.get(url, headers=self.headers)
        if r.status_code == requests.codes.ok:
            links = json.loads(r.text)["_links"]
            if link_name in links:
                return links[link_name]["href"]

    def get_schemas(self, latest_only=True, high_level_entity=None, domain_entity=None, concrete_entity=None):
        schema_url = self.get_schemas_url()
        all_schemas = []
        filtered_schemas = {}

        if latest_only:
            search_url = self._get_url_for_link(schema_url, "search")
            r = requests.get(search_url, headers=self.headers)
            if r.status_code == requests.codes.ok:
                response_j = json.loads(r.text)
                all_schemas = list(self.getRelatedEntities("latestSchemas", response_j, "schemas"))
        else:
            all_schemas = list(self.getEntities(schema_url, "schemas"))

        if high_level_entity:
            all_schemas = list(filter(lambda schema: schema.get('highLevelEntity') == high_level_entity, all_schemas))

        if domain_entity:
            all_schemas = list(filter(lambda schema: schema.get('domainEntity') == domain_entity, all_schemas))

        if concrete_entity:
            all_schemas = list(filter(lambda schema: schema.get('concreteEntity') == concrete_entity, all_schemas))

        return all_schemas

    def get_schemas_url(self):
        if "schemas" in self.ingest_api_root:
            return self.ingest_api_root["schemas"]["href"].rsplit("{")[0]
        return None

    def getSubmissions(self):
        params = {'sort': 'submissionDate,desc'}
        r = requests.get(self.ingest_api_root["submissionEnvelopes"]["href"].rsplit("{")[0], params=params,
                         headers=self.headers)
        if r.status_code == requests.codes.ok:
            return json.loads(r.text)["_embedded"]["submissionEnvelopes"]

    def getSubmissionIfModifiedSince(self, submissionId, datetimeUTC):
        submissionUrl = self.getSubmissionUri(submissionId)
        headers = self.headers

        if datetimeUTC:
            headers = {'If-Modified-Since': datetimeUTC}

        self.logger.info('headers:' + str(headers))
        r = requests.get(submissionUrl, headers=headers)

        if r.status_code == requests.codes.ok:
            submission = json.loads(r.text)
            return submission
        else:
            self.logger.error(str(r))

    def getProjects(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/projects'
        r = requests.get(submissionUrl, headers=self.headers)
        projects = []
        if r.status_code == requests.codes.ok:
            projects = json.loads(r.text)
        return projects

    def getProjectById(self, id):
        submissionUrl = self.url + '/projects/' + id
        r = requests.get(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            project = json.loads(r.text)
            return project
        else:
            raise ValueError("Project " + id + " could not be retrieved")

    def getProjectByUuid(self, uuid):
        return self.getEntityByUuid('projects', uuid)

    def getEntityByUuid(self, entity_type, uuid):
        url = self.url + f'/{entity_type}/search/findByUuid?uuid=' + uuid

        # TODO make the endpoint consistent
        if entity_type == 'submissionEnvelopes':
            url = self.url + f'/{entity_type}/search/findByUuidUuid?uuid=' + uuid

        r = requests.get(url, headers=self.headers)
        r.raise_for_status()
        return r.json()

    def getFileBySubmissionUrlAndFileName(self, submissionUrl, fileName):
        searchUrl = self._get_url_for_link(self.url + '/files/search', 'findBySubmissionEnvelopesInAndFileName')
        searchUrl = searchUrl.replace('{?submissionEnvelope,fileName}', '')
        r = requests.get(searchUrl, params={'submissionEnvelope': submissionUrl, 'fileName': fileName})
        if r.status_code == requests.codes.ok:
            return r.json()
        return None

    def getSubmissionEnvelope(self, submissionUrl):
        r = requests.get(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            submissionEnvelope = json.loads(r.text)
            return submissionEnvelope
        else:
            raise ValueError("Submission Envelope " + submissionUrl + " could not be retrieved")

    def getSubmissionByUuid(self, submissionUuid):
        searchByUuidLink = self._get_url_for_link(self.url + '/submissionEnvelopes/search', 'findByUuid')
        searchByUuidLink = searchByUuidLink.replace('{?uuid}', '')  # TODO: use a REST traverser instead of requests?
        r = requests.get(searchByUuidLink, params={'uuid': submissionUuid})

        if 200 <= r.status_code < 300:
            return r.json()
        else:
            r.raise_for_status()

    def getFiles(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/files'
        r = requests.get(submissionUrl, headers=self.headers)
        files = []
        if r.status_code == requests.codes.ok:
            files = json.loads(r.text)
        return files

    def getBundleManifests(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/bundleManifests'
        r = requests.get(submissionUrl, headers=self.headers)
        bundleManifests = []

        if r.status_code == requests.codes.ok:
            bundleManifests = json.loads(r.text)
        return bundleManifests

    def createSubmission(self, token):
        auth_headers = {
            'Content-type': 'application/json',
            'Authorization': token
        }

        try:
            r = requests.post(self.ingest_api_root["submissionEnvelopes"]["href"].rsplit("{")[0], data="{}",
                              headers=auth_headers)
            r.raise_for_status()
            submission = r.json()
            submission_url = submission["_links"]["self"]["href"].rsplit("{")[0]
            self.submission_links[submission_url] = submission["_links"]
            return submission_url
        except requests.exceptions.RequestException as err:
            self.logger.error("Request failed: " + str(err))
            raise

    def get_submission_links(self, submission_url):
        if not self.submission_links.get(submission_url):
            r = requests.get(submission_url, headers=self.headers)
            r.raise_for_status()
            self.submission_links[submission_url] = r.json()["_links"]

        return self.submission_links.get(submission_url)

    def get_link_in_submisssion(self, submission_url, link_name):
        links = self.get_submission_links(submission_url)
        link_obj = links.get(link_name) # TODO what if link doesn't exist
        link = link_obj['href'].rsplit("{")[0]

        return link

    def finishSubmission(self, submissionUrl):
        r = requests.put(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.update:
            self.logger.info("Submission complete!")
            return r.text

    def updateSubmissionState(self, submissionId, state):
        state_url = self.getSubmissionStateUrl(submissionId, state)

        if state_url:
            r = requests.put(state_url, headers=self.headers)

        return self.handleResponse(r)

    def getSubmissionStateUrl(self, submissionId, state):
        submissionUrl = self.getSubmissionUri(submissionId)
        response = requests.get(submissionUrl, headers=self.headers)
        submission = self.handleResponse(response)

        if submission and state in submission['_links']:
            return submission['_links'][state]["href"].rsplit("{")[0]

        return None

    def handleResponse(self, response):
        if response.ok:
            return json.loads(response.text)
        else:
            self.logger.error('Response:' + response.text)
            return None

    def getSubmissionUri(self, submissionId):
        return self.ingest_api_root["submissionEnvelopes"]["href"].rsplit("{")[0] + "/" + submissionId

    def get_full_url(self, callback_link):
        return urljoin(self.url, callback_link)

    def get_process(self, process_url):
        r = requests.get(process_url, headers=self.headers)
        r.raise_for_status()
        return r.json()

    def getAnalyses(self, submissionUrl):
        return self.getEntities(submissionUrl, "analyses")

    def getEntities(self, submissionUrl, entityType, pageSize=None):
        r = requests.get(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            if entityType in json.loads(r.text)["_links"]:
                if not pageSize:
                    yield from self._getAllObjectsFromSet(json.loads(r.text)["_links"][entityType]["href"], entityType)
                else:
                    yield from self._getAllObjectsFromSet(json.loads(r.text)["_links"][entityType]["href"], entityType, pageSize)

    def _getAllObjectsFromSet(self, url, entityType, pageSize=None):
        params = dict()
        if pageSize:
            params = {"size": pageSize}

        r = requests.get(url, headers=self.headers, params=params)
        r.raise_for_status()
        if r.status_code == requests.codes.ok:
            if "_embedded" in json.loads(r.text):
                for entity in json.loads(r.text)["_embedded"][entityType]:
                    yield entity
                if "next" in json.loads(r.text)["_links"]:
                    for entity2 in self._getAllObjectsFromSet(json.loads(r.text)["_links"]["next"]["href"], entityType):
                        yield entity2

    def getRelatedEntities(self, relation, entity, entityType):
        # get the self link from entity
        if relation in entity["_links"]:
            entityUri = entity["_links"][relation]["href"]
            for entity in self._getAllObjectsFromSet(entityUri, entityType):
                yield entity

    def _updateStatusToPending(self, submissionUrl):
        r = requests.patch(submissionUrl, data="{\"submissionStatus\" : \"Pending\"}", headers=self.headers)

    def createProject(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "projects", self.token)

    def createBiomaterial(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "biomaterials")

    def createProcess(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "processes")

    def createSubmissionManifest(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, 'submissionManifest')

    def createSubmissionError(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, 'submissionErrors')

    def createProtocol(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "protocols")

    def createFile(self, submissionUrl, file_name, jsonObject):
        # TODO: why do we need the submission's links before we can create a file on it?
        # TODO: submission_links should be a cache;
        # TODO: getting a submission's links should look in the cache before retrieving it from the API

        fileSubmissionsUrl = self.get_link_in_submisssion(submissionUrl, 'files')

        fileSubmissionsUrl = fileSubmissionsUrl + "/" + quote(file_name)

        fileToCreateObject = {
            "fileName": file_name,
            "content": json.loads(jsonObject)  # TODO jsonObject should be a dict()
        }

        time.sleep(0.001)
        r = requests.post(fileSubmissionsUrl, data=json.dumps(fileToCreateObject), headers=self.headers)

        # TODO Investigate why core is returning internal server error
        if r.status_code == requests.codes.conflict or r.status_code == requests.codes.internal_server_error:
            searchFiles = self.getFileBySubmissionUrlAndFileName(submissionUrl, file_name)

            if searchFiles and searchFiles.get('_embedded') and searchFiles['_embedded'].get('files'):
                fileInIngest = searchFiles['_embedded'].get('files')[0]
                content = fileInIngest.get('content')
                newContent = json.loads(jsonObject)

                if content:
                    content.update(newContent)
                else:
                    content = newContent

                fileUrl = fileInIngest['_links']['self']['href']
                time.sleep(0.001)
                r = requests.patch(fileUrl, data=json.dumps({'content': content}), headers=self.headers)
                self.logger.debug(f'Updating existing content of file {fileUrl}.')

        r.raise_for_status()

        return r.json()

    def createEntity(self, submissionUrl, jsonObject, entityType, token=None):
        auth_headers = {'Content-type': 'application/json',
                        'Authorization': token
                        }
        submissionUrl = self.get_link_in_submisssion(submissionUrl, entityType)

        self.logger.debug("posting " + submissionUrl)
        r = requests.post(submissionUrl, data=jsonObject, headers=auth_headers)
        r.raise_for_status()
        return r.json()

    # given a HCA object return the URI for the object from ingest
    def getObjectId(self, entity):
        if "_links" in entity:
            entityUrl = entity["_links"]["self"]["href"].rsplit("{")[0]
            return entityUrl
        raise ValueError('Can\'t get id for ' + json.dumps(entity) + ' is it a HCA entity?')

    def getObjectUuid(self, entityUri):
        r = requests.get(entityUri,
                         headers=self.headers)
        if r.status_code == requests.codes.ok:
            return json.loads(r.text)["uuid"]["uuid"]

    def linkEntity(self, fromEntity, toEntity, relationship):
        if not fromEntity:
            raise ValueError("Error: fromEntity is None")

        if not toEntity:
            raise ValueError("Error: toEntity is None")

        if not relationship:
            raise ValueError("Error: relationship is None")

        # check each dict in turn for non-None-ness

        fromEntityLinks = fromEntity["_links"] if "_links" in fromEntity else None
        if not fromEntityLinks:
            raise ValueError("Error: fromEntity has no _links")

        fromEntityLinksRelationship = fromEntityLinks[relationship] if relationship in fromEntityLinks else None
        if not fromEntityLinksRelationship:
            raise ValueError("Error: fromEntityLinks has no {0} relationship".format(relationship))

        fromEntityLinksRelationshipHref = fromEntityLinksRelationship["href"] if "href" in fromEntityLinksRelationship else None
        if not fromEntityLinksRelationshipHref:
            raise ValueError("Error: fromEntityLinksRelationship for relationship {0} has no href".format(relationship))

        fromUri = fromEntity["_links"][relationship]["href"]
        toUri = self.getObjectId(toEntity)

        self._retry_when_http_error(0, self._post_link_entity, fromUri, toUri)

    def _post_link_entity(self, fromUri, toUri):
        self.logger.debug('fromUri ' + fromUri + ' toUri:' + toUri);

        headers = {'Content-type': 'text/uri-list'}

        r = requests.post(fromUri.rsplit("{")[0],
                          data=toUri.rsplit("{")[0], headers=headers)

        return r

    def _retry_when_http_error(self, tries, func, *args):
        max_retries = 5

        if tries < max_retries:
            if tries > 1:
                self.logger.info("no of tries: " + str(tries + 1))

            r = None
            
            try:
                time.sleep(0.001)
                r = func(*args)
                r.raise_for_status()
            except HTTPError:
                self.logger.error("\nResponse was: " + str(r.status_code) + " (" + r.text + ")")
                tries += 1
                time.sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            except requests.ConnectionError as e:
                self.logger.exception(str(e))
                tries += 1
                time.sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            except Exception as e:
                self.logger.exception(str(e))
                tries += 1
                time.sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            return r
        else:
            error_message = "Maximum no of tries reached: " + str(max_retries)
            self.logger.error(error_message)
            return None

    def _request_post(self, url, data, params, headers):
        if params:
            return requests.post(url, data=data, params=params, headers=headers)

        return requests.post(url, data=data, headers=headers)

    def _request_put(self, url, data, params, headers):
        if params:
            return requests.put(url, data=data, params=params, headers=headers)

        return requests.put(url, data=data, headers=headers)

    def createBundleManifest(self, bundleManifest):
        r = self._retry_when_http_error(0, self._post_bundle_manifest, bundleManifest, self.ingest_api_root["bundleManifests"]["href"].rsplit("{")[0])

        if not (200 <= r.status_code < 300):
            error_message = "Failed to create bundle manifest at URL {0} with request payload: {1}".format(self.ingest_api_root["bundleManifests"]["href"].rsplit("{")[0],
                                                                                                           json.dumps(bundleManifest.__dict__))
            self.logger.error(error_message)
            raise ValueError(error_message)
        else:
            self.logger.info("successfully created bundle manifest")

    def _post_bundle_manifest(self, bundleManifest, url):
        return requests.post(url, data=json.dumps(bundleManifest.__dict__), headers=self.headers)

    def updateSubmissionWithStagingCredentials(self, subUrl, uuid, submissionCredentials):
        stagingDetails = \
            {
                "stagingDetails": {
                    "stagingAreaUuid": {
                        "uuid": uuid
                    },
                    "stagingAreaLocation": {
                        "value": submissionCredentials
                    }
                }
            }

        if self.retrySubmissionUpdateWithStagingDetails(subUrl, stagingDetails, 0):
            self.logger.debug("envelope updated with staging details " + json.dumps(stagingDetails))
        else:
            self.logger.error("Failed to update envelope with staging details: " + json.dumps(stagingDetails))

    def retrySubmissionUpdateWithStagingDetails(self, subUrl, stagingDetails, tries):
        if tries < 5:
            # do a GET request to get latest submission envelope
            entity_response = requests.get(subUrl)
            etag = entity_response.headers['ETag']
            if etag:
                # set the etag header so we get 412 if someone beats us to set validating
                self.headers['If-Match'] = etag
                r = requests.patch(subUrl, data=json.dumps(stagingDetails))
                try:
                    r.raise_for_status()
                    return True
                except HTTPError:
                    self.logger.error("PATCHing submission envelope with creds failed, retrying")
                    tries += 1
                    self.retrySubmissionUpdateWithStagingDetails(subUrl, stagingDetails, tries)
        else:
            return False


class BundleManifest:
    def __init__(self):
        self.bundleUuid = str(uuid.uuid4())
        self.envelopeUuid = {}
        self.dataFiles = []
        self.fileBiomaterialMap = {}
        self.fileProcessMap = {}
        self.fileFilesMap = {}
        self.fileProjectMap = {}
        self.fileProtocolMap = {}


from unittest import TestCase

from ingest.importer.data_node import DataNode


class DataNodeTest(TestCase):

    def test___setitem__(self):
        # given:
        node = DataNode()

        # when:
        node['path.to.node'] = 'value'
        node['path.to.nested.field'] = 347

        # then:
        dict = node.as_dict()
        self.assertEqual('value', dict['path']['to']['node'])
        self.assertEqual(347, dict['path']['to']['nested']['field'])

    def test___getitem__(self):
        # given:
        defaults = {
            'first_name': 'Juan',
            'last_name': 'dela Cruz',
            'age': 39,
            'address': {
                'city': 'Manila',
                'country': 'Philippines'
            }
        }

        # and:
        data_node = DataNode(defaults=defaults)

        # expect:
        self.assertEqual('Juan', data_node['first_name'])
        self.assertEqual(39, data_node['age'])

        # and:
        self.assertEqual('Manila', data_node['address.city'])
        self.assertEqual('Philippines', data_node['address.country'])

    def test___getitem___non_existent_path(self):
        # given:
        defaults = {'product': {'name': 'biscuit', 'id': '123'}}
        data_node = DataNode(defaults=defaults)

        # expect:
        # TODO this should probably throw exception instead, indicating path does not exist
        self.assertIsNone(data_node['product.path.does.not.exist'])
        self.assertIsNone(data_node['simply.does.not.exist'])



import json
import logging
import openpyxl

import ingest.importer.submission

from ingest.importer.conversion import template_manager
from ingest.importer.conversion.template_manager import TemplateManager
from ingest.importer.spreadsheet.ingest_workbook import IngestWorkbook
from ingest.importer.submission import IngestSubmitter, EntityMap, EntityLinker


format = '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(format=format)


class XlsImporter:

    # TODO why does the importer need to refer to an IngestApi instance?
    # Seems like it should be the IngestSubmitter that takes care of this detail
    def __init__(self, ingest_api):
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)

    def dry_run_import_file(self, file_path, project_uuid=None):
        spreadsheet_json, template_mgr = self._generate_spreadsheet_json(file_path, project_uuid)
        entity_map = self._process_links_from_spreadsheet(template_mgr, spreadsheet_json)

        return entity_map

    def _generate_spreadsheet_json(self, file_path, project_uuid=None):

        ingest_workbook = self._create_ingest_workbook(file_path)
        template_mgr = None

        try:
            template_mgr = template_manager.build(ingest_workbook.get_schemas(), self.ingest_api)
        except Exception as e:
            raise SchemaRetrievalError(
                'An error was encountered while retrieving the schema information to process the spreadsheet.')

        workbook_importer = WorkbookImporter(template_mgr)
        spreadsheet_json = workbook_importer.do_import(ingest_workbook, project_uuid)

        return spreadsheet_json, template_mgr

    def import_file(self, file_path, submission_url, project_uuid=None):
        error_json = None
        submission = None
        try:
            spreadsheet_json, template_mgr = self._generate_spreadsheet_json(file_path, project_uuid)
            entity_map = self._process_links_from_spreadsheet(template_mgr, spreadsheet_json)

            submitter = IngestSubmitter(self.ingest_api)

            # TODO the submission_url should be passed to the IngestSubmitter instead
            submission = submitter.submit(entity_map, submission_url)

        except ingest.importer.submission.Error as e:
            error_json = json.dumps({
                'errorCode': 'ingest.importer.submission',
                'errorType': 'Error',
                'message':  'An error during submission occurred.',
                'details': str(e),

            })
            self.logger.error(str(e), exc_info=True)
        except Exception as e:
            error_json = json.dumps({
                'errorCode': 'ingest.importer.error',
                'errorType': 'Error',
                'message': 'An error during submission occurred.',
                'details': str(e),
            })
            self.logger.error(str(e), exc_info=True)

        if error_json:
            self.ingest_api.createSubmissionError(submission_url, error_json)
        else:
            self.logger.info(f'Submission in {submission_url} is done!')

        return submission

    @staticmethod
    def _create_ingest_workbook(file_path):
        workbook = openpyxl.load_workbook(filename=file_path, read_only=True)
        return IngestWorkbook(workbook)

    @staticmethod
    def _process_links_from_spreadsheet(template_mgr, spreadsheet_json):
        entity_map = EntityMap.load(spreadsheet_json)
        entity_linker = EntityLinker(template_mgr)
        entity_map = entity_linker.process_links_from_spreadsheet(entity_map)
        return entity_map


class WorkbookImporter:

    def __init__(self, template_mgr):
        self.worksheet_importer = IdentifiableWorksheetImporter()
        self.template_mgr = template_mgr
        self.logger = logging.getLogger(__name__)

    def do_import(self, workbook: IngestWorkbook, project_uuid=None):
        spreadsheet_json = {}

        self.import_or_reference_project(project_uuid, spreadsheet_json, workbook)

        for worksheet in workbook.importable_worksheets():
            concrete_entity = self.template_mgr.get_concrete_entity_of_tab(worksheet.title)
            domain_entity = self.template_mgr.get_domain_entity(concrete_entity)

            entities_dict = self.worksheet_importer.do_import(worksheet, self.template_mgr)
            if spreadsheet_json.get(domain_entity) is None:
                spreadsheet_json[domain_entity] = {}

            spreadsheet_json[domain_entity].update(entities_dict)

        return spreadsheet_json

    def import_or_reference_project(self, project_uuid, spreadsheet_json, workbook):
        project_dict = None
        if not project_uuid:
            project_dict = self.import_project(workbook)
        else:
            project_dict = self._create_project_dict(project_uuid)
        spreadsheet_json['project'] = project_dict

    def import_project(self, workbook):
        project_worksheet = workbook.get_project_worksheet()
        project_importer = ProjectWorksheetImporter()

        project_dict = project_importer.do_import(project_worksheet, self.template_mgr)

        project_record = list(project_dict.values())[0]

        for worksheet in workbook.module_worksheets():
            if worksheet:
                module_importer = ModuleWorksheetImporter('project', workbook.get_module_field(worksheet.title))
                records = module_importer.do_import(worksheet, self.template_mgr)
                field_name = module_importer.property
                project_record['content'][field_name] = list(
                    map(lambda record: record['content'][field_name][0], records))

        return project_dict

    def _create_project_dict(self, project_id):
        project_dict = {}
        project_dict[project_id] = {}
        project_dict[project_id]['is_reference'] = True

        return project_dict


class WorksheetImporter:

    KEY_HEADER_ROW_IDX = 4
    USER_FRIENDLY_HEADER_ROW_IDX = 2
    START_ROW_IDX = 5

    UNKNOWN_ID_PREFIX = '_unknown_'

    def __init__(self):
        self.unknown_id_ctr = 0
        self.logger = logging.getLogger(__name__)
        self.concrete_entity = None

    def do_import(self, worksheet, template: TemplateManager):
        row_template = template.create_row_template(worksheet)
        self.concrete_entity = template.get_concrete_entity_of_tab(worksheet.title)
        return self._import_using_row_template(template, worksheet, row_template)

    def _import_using_row_template(self, template, worksheet, row_template):
        records = {}
        rows = self._get_data_rows(worksheet, template)
        for index, row in enumerate(rows):
            metadata = row_template.do_import(row)

            record_id = self._determine_record_id(metadata)

            records[record_id] = {
                'content': metadata.content.as_dict(),
                'links_by_entity': metadata.links,
                'external_links_by_entity': metadata.external_links,
                'linking_details': metadata.linking_details,
                'concrete_type': self.concrete_entity
            }
        return records

    @staticmethod
    def _is_empty_row(row):
        return all(cell.value is None for cell in row)

    def _get_data_rows(self, worksheet, template):
        header_row = template.get_header_row(worksheet)
        max_row = self._compute_max_row(worksheet) - self.START_ROW_IDX
        rows = worksheet.iter_rows(row_offset=self.START_ROW_IDX, max_row=max_row)
        return [row[:len(header_row)] for row in rows if not WorksheetImporter._is_empty_row(row)]

    # NOTE: there are no tests around this because it's too complicated to setup the
    # scenario where the worksheet returns an erroneous max_row value.
    @staticmethod
    def _compute_max_row(worksheet):
        max_row = worksheet.max_row
        if max_row is None:
            worksheet.calculate_dimension(force=True)
            max_row = worksheet.max_row
        return max_row

    def _determine_record_id(self, metadata):
        record_id = metadata.object_id

        if record_id is None:
            record_id = self._generate_id()

        return record_id

    def _generate_id(self):
        self.unknown_id_ctr = self.unknown_id_ctr + 1
        return f'{self.UNKNOWN_ID_PREFIX}{self.unknown_id_ctr}'


class IdentifiableWorksheetImporter(WorksheetImporter):

    def do_import(self, worksheet, template: TemplateManager):
        records = super(IdentifiableWorksheetImporter, self).do_import(worksheet, template)

        if not self.concrete_entity:
            raise InvalidTabName(worksheet.title)

        if self.unknown_id_ctr:
            raise RowIdNotFound(worksheet.title)

        return records


class ProjectWorksheetImporter(WorksheetImporter):

    def do_import(self, worksheet, template: TemplateManager):
        records = super(ProjectWorksheetImporter, self).do_import(worksheet, template)

        if len(records.keys()) == 0:
            raise NoProjectFound()

        if len(records.keys()) > 1:
            raise MultipleProjectsFound()

        return records


class ModuleWorksheetImporter(WorksheetImporter):
    def __init__(self, parent_entity, property):
        super(ModuleWorksheetImporter, self).__init__()
        self.parent_entity = parent_entity
        self.property = property

    def do_import(self, worksheet, template: TemplateManager):
        row_template = template.create_simple_row_template(worksheet)
        records = self._import_using_row_template(template, worksheet, row_template)
        return list(records.values())


class MultipleProjectsFound(Exception):
    def __init__(self):
        message = f'The spreadsheet should only be associated to a single project.'
        super(MultipleProjectsFound, self).__init__(message)


class NoProjectFound(Exception):
    def __init__(self):
        message = f'The spreadsheet should be associated to a project.'
        super(NoProjectFound, self).__init__(message)


class RowIdNotFound(Exception):
    def __init__(self, tab_name):
        message = f'No identifier was found for some rows in "{tab_name}" tab.'
        super(RowIdNotFound, self).__init__(message)
        self.tab_name = tab_name


class SchemaRetrievalError(Exception):
    pass


class InvalidTabName(Exception):
    def __init__(self, tab_name):
        message = f'The {tab_name} tab does not correspond to any entity in metadata schema.'
        super(InvalidTabName, self).__init__(message)
        self.tab_name = tab_name




from unittest import TestCase
from unittest.mock import patch

from mock import MagicMock

from ingest.api.ingestapi import IngestApi

import json

mock_ingest_api_url = "http://mockingestapi.com"
mock_submission_envelope_id = "mock-envelope-id"


class IngestApiTest(TestCase):

    def test_create_file(self):
        api_url = mock_ingest_api_url
        submission_id = mock_submission_envelope_id
        submission_url = api_url + "/" + submission_id
        filename = "mock-filename"
        file_create_url = submission_url + "/files/" + filename

        # mock the load_root()
        with patch('ingest.api.ingestapi.IngestApi.get_root_url') as mock_load_root:
            root_links = dict()
            root_links["file"] = {"href": api_url + "/files"}
            root_links["submissionEnvelopes"] = {"href": api_url + "/submissionEnvelopes"}
            mock_load_root.return_value = root_links

            ingestapi = IngestApi(api_url)
            ingestapi.submission_links[submission_url] = {
                'files': {
                    'href': submission_url + "/files"
                }
            }

            with patch('ingest.api.ingestapi.requests.post') as mock_post:

                def mock_post_side_effect(*args, **kwargs):
                    mock_response = {}
                    mock_response['json'] = lambda _self: {}
                    mock_response['text'] = "{}"
                    mock_response['raise_for_status'] = MagicMock()

                    if args[0] == file_create_url:
                        mock_response['status_code'] = 201
                    else:
                        mock_response['status_code'] = 404

                    return type("MockResponse", (), mock_response)()

                mock_post.side_effect = mock_post_side_effect
                ingestapi.createFile(submission_url, filename, "{}")

    def test_get_submission_by_uuid(self):
        api_url = mock_ingest_api_url
        mock_submission_uuid = "mock-submission-uuid"
        submissions_url = api_url + "/submissionEnvelopes"
        submission_search_uri = submissions_url + "/search"
        findByUuidRel = "findByUuid"
        findByUuidHref = submission_search_uri + "/findByUuidHref{?uuid}"

        ingestapi = IngestApi(api_url, dict())
        with patch('ingest.api.ingestapi.IngestApi._get_url_for_link') as mock_get_url_for_link:
            def mock_get_url_for_link_patch(*args, **kwargs):
                if args[0] == submission_search_uri and args[1] == findByUuidRel:
                    return findByUuidHref

            mock_get_url_for_link.side_effect = mock_get_url_for_link_patch

            with patch('ingest.api.ingestapi.requests.get') as mock_requests_get:
                def mock_get_side_effect(*args, **kwargs):
                    mock_response = {}
                    mock_response_payload = {}

                    if args[0] == submission_search_uri + "/findByUuidHref" \
                            and 'params' in kwargs \
                            and 'uuid' in kwargs['params'] \
                            and kwargs['params']['uuid'] == mock_submission_uuid:
                        mock_response['status_code'] = 200
                        mock_response_payload = {"uuid": {"uuid": mock_submission_uuid}}
                    else:
                        mock_response['status_code'] = 404

                    mock_response['json'] = lambda _self: mock_response_payload
                    mock_response['text'] = json.dumps(mock_response_payload)

                    def raise_for_status():
                        raise Exception("test failed")

                    mock_response['raise_for_status'] = lambda _self: raise_for_status()

                    return type("MockResponse", (), mock_response)()

                mock_requests_get.side_effect = mock_get_side_effect

                assert 'uuid' in ingestapi.getSubmissionByUuid(mock_submission_uuid)


from enum import Enum

from ingest.importer.conversion import utils, data_converter
from ingest.importer.conversion.data_converter import DataType, CONVERTER_MAP, ListConverter


class ConversionType(Enum):
    UNDEFINED = 0,
    MEMBER_FIELD = 1,
    FIELD_OF_LIST_ELEMENT = 2,
    IDENTITY = 3,
    LINKED_IDENTITY = 4,
    EXTERNAL_REFERENCE = 5,
    LINKING_DETAIL = 6


class ColumnSpecification:

    def __init__(self, field_name, object_type, main_category, data_type, multivalue=False,
                 multivalue_parent=False, identity: bool=False, external_reference: bool=False,
                 order_of_occurence: int = 1):
        self.field_name = field_name
        self.object_type = object_type
        self.main_category = main_category
        self.data_type = data_type
        self.multivalue = multivalue
        self.multivalue_parent = multivalue_parent
        self.identity = identity
        self.external_reference = external_reference
        self.order_of_occurence = order_of_occurence

    def is_multivalue(self):
        return self.multivalue

    def is_field_of_list_element(self):
        return self.multivalue_parent

    def is_identity(self):
        return self.identity

    def is_external_reference(self):
        return self.external_reference

    def get_conversion_type(self):
        if self._represents_an_object_field():
            conversion_type = self._determine_conversion_type_for_object_field()
        else:
            conversion_type = self._determine_conversion_type_for_reference_field()
        return conversion_type

    def _represents_an_object_field(self):
        entity_type = utils.extract_root_field(self.field_name)
        return entity_type == self.object_type and self.order_of_occurence == 1

    def _determine_conversion_type_for_object_field(self):
        if self.identity:
            conversion_type = ConversionType.IDENTITY
        elif self.multivalue_parent:
            conversion_type = ConversionType.FIELD_OF_LIST_ELEMENT
        else:
            conversion_type = ConversionType.MEMBER_FIELD
        return conversion_type

    def _determine_conversion_type_for_reference_field(self):
        if self.identity:
            if self.external_reference:
                conversion_type = ConversionType.EXTERNAL_REFERENCE
            else:
                conversion_type = ConversionType.LINKED_IDENTITY
        else:
            conversion_type = ConversionType.LINKING_DETAIL
        return conversion_type

    def determine_converter(self):
        if not self.multivalue:
            converter = CONVERTER_MAP.get(self.data_type, data_converter.DEFAULT)
        else:
            converter = ListConverter(self.data_type)
        return converter

    @staticmethod
    def build_raw(field_name, object_type, main_category, raw_spec, order_of_occurence=1, parent=None):
        data_type = DataType.find(raw_spec.get('value_type'))
        multivalue = bool(raw_spec.get('multivalue'))
        multivalue_parent = bool(parent.get('multivalue')) if parent != None else False
        identity: bool = bool(raw_spec.get('identifiable'))
        external_reference = bool(raw_spec.get('external_reference'))
        return ColumnSpecification(field_name, object_type, main_category, data_type,
                                   multivalue=multivalue, multivalue_parent=multivalue_parent,
                                   identity=identity, external_reference=external_reference,
                                   order_of_occurence=order_of_occurence)



from unittest import TestCase

from mock import MagicMock, patch, call
from openpyxl import Workbook

from ingest.importer.conversion import conversion_strategy
from ingest.importer.conversion.column_specification import ColumnSpecification
from ingest.importer.conversion.conversion_strategy import CellConversion
from ingest.importer.conversion.template_manager import TemplateManager, RowTemplate
from ingest.importer.data_node import DataNode


def _mock_schema_template_lookup(value_type='string', multivalue=False):
    schema_template = MagicMock(name='schema_template')
    single_string_spec = {
        'value_type': value_type,
        'multivalue': multivalue
    }
    schema_template.lookup = MagicMock(name='lookup', return_value=single_string_spec)
    return schema_template


class TemplateManagerTest(TestCase):

    def test_create_template_node(self):
        # given:
        tab_config = MagicMock(name='tab_config')
        tab_config.get_key_for_label = MagicMock(return_value='concrete_entity')

        schema_template = MagicMock(name='schema_template')
        schema_template.get_tabs_config = MagicMock(return_value=tab_config)

        schema_url = 'https://schema.humancellatlas.org/type/biomaterial/5.1.0/donor_organsim'

        lookup_map = {
            'concrete_entity': {
                'schema': {
                    'domain_entity': 'biomaterial',
                    'url': schema_url
                }
            }
        }

        schema_template.lookup = lambda key: lookup_map.get(key)

        ingest_api = MagicMock(name='ingest_api')

        # and:
        template_manager = TemplateManager(schema_template, ingest_api)

        # and:
        workbook = Workbook()
        donor_worksheet = workbook.create_sheet('Donor')

        # when:
        data_node:DataNode = template_manager.create_template_node(donor_worksheet)

        # then:
        data = data_node.as_dict()
        self.assertEqual(schema_url, data.get('describedBy'))
        self.assertEqual('biomaterial', data.get('schema_type'))

    @patch.object(ColumnSpecification, 'build_raw')
    @patch.object(conversion_strategy, 'determine_strategy')
    def test_create_row_template(self, determine_strategy, build_raw):
        # given:
        schema_template = MagicMock(name='schema_template')
        ingest_api = MagicMock(name='ingest_api')

        # and:
        tabs_config = MagicMock('tabs_config')
        object_type = 'sample_object'
        tabs_config.get_key_for_label = MagicMock(return_value=object_type)
        schema_template.get_tabs_config = MagicMock(return_value=tabs_config)

        # and: set up column spec
        name_column_spec = MagicMock('name_column_spec')
        numbers_column_spec = MagicMock('numbers_column_spec')
        build_raw.side_effect = [name_column_spec, numbers_column_spec]

        # and: set up raw spec
        name_raw_spec = MagicMock('name_raw_spec')
        name_raw_parent_spec = MagicMock('name_raw_parent_spec')
        numbers_raw_spec = MagicMock('numbers_raw_spec')

        # TODO move the logic of creating the column spec to SchemaTemplate
        # and:
        schema = {'schema': {'domain_entity': 'main_category/subdomain'}}
        spec_map = {
            'user.profile.first_name': name_raw_spec,
            'user.profile': name_raw_parent_spec,
            'numbers': numbers_raw_spec,
            'user': schema
        }
        schema_template.lookup = lambda key: spec_map.get(key, None)

        # and:
        name_strategy = MagicMock('name_strategy')
        numbers_strategy = MagicMock('numbers_strategy')
        determine_strategy.side_effect = [name_strategy, numbers_strategy]

        # and: prepare worksheet
        workbook = Workbook()
        worksheet = workbook.create_sheet('sample')
        worksheet['A4'] = 'user.profile.first_name'
        worksheet['B4'] = 'numbers'

        # when:
        template_manager = TemplateManager(schema_template, ingest_api)
        row_template: RowTemplate = template_manager.create_row_template(worksheet)

        # then:
        expected_calls = [
            call('user.profile.first_name', 'sample_object', 'main_category', name_raw_spec,
                 order_of_occurence=1, parent=name_raw_parent_spec),
            call('numbers', 'sample_object', None, numbers_raw_spec, order_of_occurence=1, parent=None)
        ]
        build_raw.assert_has_calls(expected_calls)
        determine_strategy.assert_has_calls([call(name_column_spec), call(numbers_column_spec)])

        # and:
        self.assertIsNotNone(row_template)
        self.assertEqual(2, len(row_template.cell_conversions))
        self.assertTrue(name_strategy in row_template.cell_conversions)
        self.assertTrue(numbers_strategy in row_template.cell_conversions)

    @patch.object(ColumnSpecification, 'build_raw')
    @patch.object(conversion_strategy, 'determine_strategy')
    def test_create_row_template_with_default_values(self, determine_strategy, build_raw):
        # given:
        schema_template = MagicMock('schema_template')
        ingest_api = MagicMock(name='ingest_api')

        # and:
        schema_url = 'http://schema.sample.com/profile'
        self._mock_schema_lookup(schema_template, schema_url=schema_url, main_category='profile',
                                 object_type='profile_type')

        # and:
        build_raw.return_value = MagicMock('column_spec')
        determine_strategy.return_value = FakeConversion('')

        # and:
        workbook = Workbook()
        worksheet = workbook.create_sheet('profile')
        worksheet['A4'] = 'profile.name'

        # when:
        template_manager = TemplateManager(schema_template, ingest_api)
        template_manager.get_schema_url = MagicMock(return_value=schema_url)
        row_template = template_manager.create_row_template(worksheet)

        # then:
        content_defaults = row_template.default_values
        self.assertIsNotNone(content_defaults)
        self.assertEqual(schema_url, content_defaults.get('describedBy'))
        self.assertEqual('profile', content_defaults.get('schema_type'))

    @patch.object(conversion_strategy, 'determine_strategy')
    def test_create_row_template_with_none_header(self, determine_strategy):
        # given:
        schema_template = MagicMock('schema_template')
        ingest_api = MagicMock(name='ingest_api')

        # and:
        do_nothing_strategy = FakeConversion('')
        determine_strategy.return_value = do_nothing_strategy

        # and:
        self._mock_schema_lookup(schema_template)

        # and:
        workbook = Workbook()
        worksheet = workbook.create_sheet('sample')
        worksheet['A4'] = None

        # when:
        template_manager = TemplateManager(schema_template, ingest_api)
        row_template = template_manager.create_row_template(worksheet)

        # then:
        self.assertEqual(0, len(row_template.cell_conversions))

    @staticmethod
    def _mock_schema_lookup(schema_template, schema_url='', object_type='', main_category=None):
        tabs_config = MagicMock('tabs_config')
        tabs_config.get_key_for_label = MagicMock(return_value=object_type)
        schema_template.get_tabs_config = MagicMock(return_value=tabs_config)
        schema_template.get_latest_schema = MagicMock(return_value=schema_url)

        domain_entity = f'{main_category}/{object_type}' if main_category else object_type
        schema = {'schema': {
            'domain_entity': domain_entity,
            'url': schema_url
        }}
        spec_map = {
            object_type: schema
        }
        schema_template.lookup = lambda key: spec_map.get(key)

    def test_get_schema_type(self):
        # given
        schema_template = MagicMock(name='schema_template')
        ingest_api = MagicMock(name='ingest_api')

        spec = {
            'schema': {
                'high_level_entity': 'type',
                'domain_entity': 'biomaterial',
                'module': 'donor_organism',
                'url': 'https://schema.humancellatlas.org/type/biomaterial/5.0.0/donor_organism'
            }
        }
        schema_template.lookup = MagicMock(name='lookup', return_value=spec)
        template_manager = TemplateManager(schema_template, ingest_api)

        # when:
        domain_entity = template_manager.get_domain_entity('cell_suspension')

        # then:
        self.assertEqual('biomaterial', domain_entity)

    def test_get_schema_url(self):
        # given
        schema_template = MagicMock(name='schema_template')
        ingest_api = MagicMock(name='ingest_api')
        latest_url = 'https://schema.humancellatlas.org/type/biomaterial/5.0.0/donor_organism'

        spec = {
            'schema': {
                'high_level_entity': 'type',
                'domain_entity': 'biomaterial',
                'module': 'donor_organism',
                'url': latest_url
            }
        }

        schema_template.lookup = MagicMock(name='lookup', return_value=spec)
        template_manager = TemplateManager(schema_template, ingest_api)
        template_manager.get_latest_schema_url = MagicMock(return_value=latest_url)

        # when:
        url = template_manager.get_schema_url('cell_suspension')

        # then:
        self.assertEqual('https://schema.humancellatlas.org/type/biomaterial/5.0.0/donor_organism', url)

    def test_get_entity_of_tab(self):
        # given

        key_label_map = {
            'Project': 'project',
            'Donor': 'donor',
            'Specimen from organism': 'specimen_from_organism'
        }

        fake_tabs_config = MagicMock(name='tabs_config')
        fake_tabs_config.get_key_for_label = lambda key: key_label_map.get(key)

        schema_template = MagicMock(name='schema_template')
        schema_template.get_tabs_config = MagicMock(return_value=fake_tabs_config)

        ingest_api = MagicMock(name='ingest_api')

        template_manager = TemplateManager(schema_template, ingest_api)

        # when:
        entity = template_manager.get_concrete_entity_of_tab('Specimen from organism')

        # then:
        self.assertEqual('specimen_from_organism', entity)


class FakeConversion(CellConversion):

    def __init__(self, field):
        self.field = field

    def apply(self, metadata, cell_data):
        metadata.define_content(self.field, cell_data)


class RowTemplateTest(TestCase):

    def test_do_import(self):
        # given:
        cell_conversions = [FakeConversion('first_name'), FakeConversion('last_name'),
                            FakeConversion('address.city'), FakeConversion('address.country')]
        row_template = RowTemplate(cell_conversions)

        # and:
        workbook = Workbook()
        worksheet = workbook.create_sheet('profile')
        worksheet['A1'] = 'Juan'
        worksheet['B1'] = 'dela Cruz'
        worksheet['C1'] = 'Manila'
        worksheet['D1'] = 'Philippines'
        row = list(worksheet.rows)[0]

        # when:
        result = row_template.do_import(row)

        # then:
        self.assertIsNotNone(result)
        self.assertEqual('Juan', result.get_content('first_name'))
        self.assertEqual('dela Cruz', result.get_content('last_name'))

        # and:
        address = result.get_content('address')
        self.assertIsNotNone(address)
        self.assertEqual('Manila', address.get('city'))
        self.assertEqual('Philippines', address.get('country'))

    def test_do_import_with_default_values(self):
        # given:
        schema_url = 'https://sample.com/list_of_things'
        default_values = {
            'describedBy': schema_url,
            'extra_field': 'extra field'
        }

        # and:
        cell_conversions = [FakeConversion('name'), FakeConversion('description')]
        row_template = RowTemplate(cell_conversions, default_values=default_values)

        # and:
        workbook = Workbook()
        worksheet = workbook.create_sheet('list_of_things')
        worksheet['A1'] = 'pen'
        worksheet['B1'] = 'a thing used for writing'
        row = list(worksheet.rows)[0]

        # when:
        result = row_template.do_import(row)

        # then:
        self.assertEqual(schema_url, result.get_content('describedBy'))
        self.assertEqual('extra field', result.get_content('extra_field'))



import os
from setuptools import setup, find_packages

base_dir = os.path.dirname(__file__)
install_requires = [line.rstrip() for line in open(os.path.join(base_dir, 'requirements.txt'))]

setup(
    name = 'hca_ingest',
    version = '0.6',
    packages = find_packages(exclude=['tests', 'tests.*']),
    install_requires = install_requires,
    include_package_data = True
)



#!/usr/bin/env python
"""
Provides mock schemas for building intsances of schema_template.py
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "25/05/2018"

from ingest.template.schema_template import SchemaTemplate
from unittest.mock import patch, MagicMock


@patch('urllib.request.urlopen')
def get_template_for_json(mock_urlopen, data="{}"):
    cm = MagicMock()
    cm.getcode.return_value = 200
    cm.read.return_value = data.encode()
    cm.__enter__.return_value = cm
    mock_urlopen.return_value = cm

    return SchemaTemplate(list_of_schema_urls=['test_url'])


import time
from unittest import TestCase
from mock import MagicMock

from ingest.utils.token_manager import TokenManager, Token


class TestTokenManager(TestCase):
    def test_get_token(self):
        token_client = MagicMock()
        token_client.retrieve_token = MagicMock(return_value='token')
        token_manager = TokenManager(token_client=token_client)
        token = token_manager.get_token()
        self.assertTrue(token)
        self.assertEqual(token, 'token')

    def test_get_token_when_expired(self):
        token_client = MagicMock()
        token_client.retrieve_token = MagicMock(return_value='token_1')
        token_manager = TokenManager(token_client=token_client)
        expired_token = token_manager.get_token()
        token_manager.token.is_expired = MagicMock(return_value=True)
        token_client.retrieve_token = MagicMock(return_value='token_2')
        new_token = token_manager.get_token()
        self.assertEqual(new_token, 'token_2')

    def test_valid_token(self):
        token = Token(value='token',
                      token_duration=3600 * 1000,
                      refresh_period=60 * 20 * 1000)
        self.assertFalse(token.is_expired())

    def test_expired_token(self):
        token = Token(value='token',
                      token_duration=1000,
                      refresh_period=0)
        time.sleep(1)
        self.assertTrue(token.is_expired())

    def test_expired_token_outside_refresh_period(self):
        token = Token(value='token',
                      token_duration=4000,
                      refresh_period=1000)
        time.sleep(3.5)
        self.assertTrue(token.is_expired())

    def test_valid_token_within_refresh_period(self):
        token = Token(value='token',
                      token_duration=4000,
                      refresh_period=1000)
        time.sleep(2)
        self.assertFalse(token.is_expired())



from unittest import TestCase

from mock import MagicMock

from ingest.importer.conversion import conversion_strategy
from ingest.importer.conversion.column_specification import ColumnSpecification, ConversionType
from ingest.importer.conversion.conversion_strategy import DirectCellConversion, \
    ListElementCellConversion, CellConversion, IdentityCellConversion, LinkedIdentityCellConversion, \
    DoNothing, ExternalReferenceCellConversion, LinkingDetailCellConversion, \
    FieldOfSingleElementListCellConversion
from ingest.importer.conversion.data_converter import StringConverter, ListConverter
from ingest.importer.conversion.exceptions import UnknownMainCategory
from ingest.importer.conversion.metadata_entity import MetadataEntity


def _mock_column_spec(field_name='field_name', main_category=None, converter=None,
                      conversion_type=ConversionType.UNDEFINED):
    column_spec: ColumnSpecification = MagicMock('column_spec')
    column_spec.field_name = field_name
    column_spec.main_category = main_category
    column_spec.determine_converter = lambda: converter
    column_spec.get_conversion_type = lambda: conversion_type
    return column_spec


class ModuleTest(TestCase):

    def test_determine_strategy_for_direct_conversion(self):
        # expect:
        self._assert_correct_strategy(ConversionType.MEMBER_FIELD, DirectCellConversion)

    def test_determine_strategy_for_field_of_list_element(self):
        # expect:
        self._assert_correct_strategy(ConversionType.FIELD_OF_LIST_ELEMENT,
                                      ListElementCellConversion,
                                      expected_converter_type=ListConverter)

    def test_determine_strategy_for_identity_field(self):
        # expect:
        self._assert_correct_strategy(ConversionType.IDENTITY, IdentityCellConversion)

    def test_determine_strategy_for_linking_detail_field(self):
        # expect
        self._assert_correct_strategy(ConversionType.LINKING_DETAIL, LinkingDetailCellConversion)

    def test_determine_strategy_for_linked_identity_field(self):
        # expect:
        self._assert_correct_strategy(ConversionType.LINKED_IDENTITY, LinkedIdentityCellConversion,
                                      expected_converter_type=ListConverter,
                                      and_also=self._assert_correct_main_category)

    def test_determine_strategy_for_external_reference_field(self):
        self._assert_correct_strategy(ConversionType.EXTERNAL_REFERENCE,
                                      ExternalReferenceCellConversion,
                                      expected_converter_type=ListConverter,
                                      and_also=self._assert_correct_main_category)

    def _assert_correct_main_category(self, strategy: CellConversion):
        self.assertEqual('product_type', strategy.main_category)

    def _assert_correct_strategy(self, conversion_type, strategy_class,
                                 expected_converter_type=None, and_also=None):
        # given:
        converter = MagicMock('converter')
        column_spec = _mock_column_spec(field_name='product.product_id',
                                        main_category='product_type', converter=converter,
                                        conversion_type=conversion_type)

        # when:
        strategy: CellConversion = conversion_strategy.determine_strategy(column_spec)

        # then:
        self.assertIsInstance(strategy, strategy_class)
        self.assertEqual('product.product_id', strategy.field)

        # and:
        if expected_converter_type is None:
            self.assertEqual(converter, strategy.converter)
        else:
            self.assertIsInstance(strategy.converter, expected_converter_type)

        # and:
        if and_also is not None:
            and_also(strategy)

    def test_determine_strategy_for_unknown_type(self):
        # given:
        converter = MagicMock('converter')
        column_spec = _mock_column_spec(field_name='product.product_id', converter=converter,
                                        conversion_type=ConversionType.UNDEFINED)

        # when:
        undefined_strategy: CellConversion = conversion_strategy.determine_strategy(column_spec)
        none_strategy: CellConversion = conversion_strategy.determine_strategy(None)

        # then:
        self.assertIsInstance(undefined_strategy, DoNothing)
        self.assertIsInstance(none_strategy, DoNothing)


class DirectCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        int_converter = MagicMock('int_converter')
        int_converter.convert = lambda __: 27

        # and:
        cell_conversion = DirectCellConversion('profile.user.age', int_converter)

        # when:
        metadata = MetadataEntity()
        cell_conversion.apply(metadata, '27')

        # then:
        self.assertEqual(27, metadata.get_content('user.age'))

    def test_apply_skips_none_data(self):
        # given:
        string_converter = StringConverter()
        cell_conversion = DirectCellConversion('product.id', string_converter)

        # when:
        metadata = MetadataEntity(content={
            'product': {
                'name': 'product name'
            }
        })
        cell_conversion.apply(metadata, None)

        # then:
        product = metadata.get_content('product')
        self.assertIsNotNone(product)
        self.assertTrue('id' not in product, '[id] not expected to be in product field')


def _create_mock_string_converter():
    converter = MagicMock('converter')
    converter.convert = lambda data: f'{data} - converted'
    return converter


class ListElementCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = ListElementCellConversion('stuff.list_of_things.name', converter)

        # when:
        metadata = MetadataEntity()
        cell_conversion.apply(metadata, 'sample')

        # then:
        list_of_things = metadata.get_content('list_of_things')
        self.assertIsNotNone(list_of_things)
        self.assertEqual(1, len(list_of_things))

        # and:
        thing = list_of_things[0]
        self.assertEqual('sample - converted', thing.get('name'))

    def test_apply_previously_processed_field(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = ListElementCellConversion('shop.user.basket.product_name', converter)

        # and:
        metadata = MetadataEntity()
        metadata.define_content('user.basket', [{'quantity': 3}])

        # when:
        cell_conversion.apply(metadata, 'apple')

        # then:
        basket = metadata.get_content('user.basket')
        self.assertEqual(1, len(basket))

        # and:
        current_element = basket[0]
        self.assertEqual(3, current_element.get('quantity'))
        self.assertEqual('apple - converted', current_element.get('product_name'))

    def test_apply_none_data(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = ListElementCellConversion('user.name', converter)

        # and:
        metadata = MetadataEntity(content={
            'user': [{'id': '65fd8'}]
        })

        # when:
        cell_conversion.apply(metadata, None)

        # then:
        list_element = metadata.get_content('user')[0]
        self.assertTrue('name' not in list_element.keys(), '[name] should not be added to element.')

    def test_apply_multiple(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = ListElementCellConversion('group.members.name', converter)

        # when:
        metadata = MetadataEntity()
        cell_conversion.apply(metadata, 'Juan||Pedro||Jane')

        # then:
        members = metadata.get_content('members')
        self.assertEqual(3, len(members))

        # and:
        expected_names = ['Juan - converted', 'Pedro - converted', 'Jane - converted']
        for member in members:
            member_name = member.get('name')
            self.assertIn(member_name, expected_names)

    def test_apply_multiple_with_previous_values(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = ListElementCellConversion('staff.developers.name', converter)

        # and:
        metadata = MetadataEntity()
        metadata.define_content('developers', [{'age': 32}, {'age': 23}])

        # when:
        cell_conversion.apply(metadata, 'John||Paul||George||Ringo')

        # then:
        developers = metadata.get_content('developers')
        self.assertEqual(4, len(developers))

        # and:
        expected_names = ['John - converted', 'Paul - converted', 'George - converted',
                          'Ringo - converted']
        expected_ages = [32, 23, None, None]
        for index, developer in enumerate(developers):
            self.assertEqual(expected_names[index], developer.get('name'))
            self.assertEqual(expected_ages[index], developer.get('age'))


class FieldOfSingleElementListCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = FieldOfSingleElementListCellConversion('library.books.title', converter)

        # when:
        metadata = MetadataEntity()
        cell_conversion.apply(metadata, 'Harry Potter')

        # then:
        books = metadata.get_content('books')
        self.assertIsNotNone(books)
        self.assertEqual(1, len(books))

        # and:
        thing = books[0]
        self.assertEqual('Harry Potter - converted', thing.get('title'))

    def test_apply_multiple_value_field(self):
        # given:
        converter = ListConverter(base_converter=_create_mock_string_converter())
        cell_conversion = FieldOfSingleElementListCellConversion('library.books.authors', converter)

        # and:
        metadata = MetadataEntity(content={
            'books': [
                {'title': 'Once Upon a Time'}
            ]
        })

        # when:
        cell_conversion.apply(metadata, 'J dela Cruz||P Smith')

        # then:
        books = metadata.get_content('books')
        self.assertEqual(1, len(books))

        # and:
        book = books[0]
        self.assertEqual(['J dela Cruz - converted', 'P Smith - converted'], book.get('authors'))


class IdentityCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = IdentityCellConversion('product.product_id', converter)

        # and:
        metadata = MetadataEntity()

        # when:
        cell_conversion.apply(metadata, 'product_no_144')

        # then:
        expected_id = 'product_no_144 - converted'
        self.assertEqual(expected_id, metadata.object_id)

        # and: identity value should be in content
        self.assertEqual(expected_id, metadata.get_content('product_id'))


class LinkedIdentityCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        cell_conversion = LinkedIdentityCellConversion('item.item_id', 'item_type')

        # and:
        metadata = MetadataEntity()

        # when:
        cell_conversion.apply(metadata, 'item_no_29')
        cell_conversion.apply(metadata, 'item_no_31||item_no_50')

        # then:
        item_types = metadata.get_links('item_type')
        self.assertEqual(3, len(item_types))

        # and:
        expected_items = [f'item_no_{number}' for number in [29, 31, 50]]
        for expected_item in expected_items:
            self.assertTrue(expected_item in item_types, f'[{expected_item}] not in list.')

    def test_apply_with_previous_entry(self):
        # given:
        cell_conversion = LinkedIdentityCellConversion('item.item_number', 'line_order')

        # and:
        metadata = MetadataEntity()
        items = ['item_no_56', 'item_no_199']
        metadata.add_links('line_order', items)

        # when:
        cell_conversion.apply(metadata, 'item_no_721')

        # then:
        actual_items = metadata.get_links('line_order')
        self.assertEqual(3, len(actual_items))

        # and:
        expected_ids = [id for id in items]
        expected_ids.append('item_no_721')
        for expected_id in expected_ids:
            self.assertTrue(expected_id in actual_items, f'[{expected_id}] not in list.')

    def test_apply_no_main_category(self):
        # given:
        cell_conversion = LinkedIdentityCellConversion('product.name', None)

        # when:
        exception_thrown = False
        try:
            cell_conversion.apply(MetadataEntity(), 'sample')
        except UnknownMainCategory:
            exception_thrown = True

        # then:
        self.assertTrue(exception_thrown, f'[{UnknownMainCategory.__name__}] not raised.')


class ExternalReferenceCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        cell_conversion = ExternalReferenceCellConversion('user.uuid', 'account')

        # when:
        metadata = MetadataEntity()
        cell_conversion.apply(metadata, '621bfa0')

        # then:
        account_list = metadata.get_external_links('account')
        self.assertIsNotNone(account_list, '[account] list in external links expected.')
        self.assertEqual(1, len(account_list))
        self.assertTrue('621bfa0' in account_list, 'Expected content not in list.')

    def test_apply_multiple_values(self):
        # given:
        cell_conversion = ExternalReferenceCellConversion('company.uuid', 'organisation')

        # when:
        metadata = MetadataEntity()
        cell_conversion.apply(metadata, '7e56de9||2fe9eb0')

        # then:
        expected_ids = ['7e56de9', '2fe9eb0']
        self.assertCountEqual(expected_ids, metadata.get_external_links('organisation'))

    def test_apply_with_previous_entries(self):
        # given:
        metadata = MetadataEntity()
        metadata.add_external_links('store_item', ['109bdd9', 'c3c35e6'])

        # and:
        cell_conversion = ExternalReferenceCellConversion('product.uuid', 'store_item')

        # when:
        cell_conversion.apply(metadata, '73de901')

        # then:
        store_item_list = metadata.get_external_links('store_item')
        self.assertIsNotNone(store_item_list, '[store_item] list in external links expected.')

        # and:
        expected_ids = ['109bdd9', '73de901', 'c3c35e6']
        self.assertCountEqual(expected_ids, store_item_list)


class LinkingDetailCellConversionTest(TestCase):

    def test_apply(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = LinkingDetailCellConversion('profile.user.name', converter)
        metadata = MetadataEntity()

        # when:
        cell_conversion.apply(metadata, 'John Doe')

        # then:
        self.assertEqual('John Doe - converted', metadata.get_linking_detail('user.name'))

    def test_apply_with_previous_entries(self):
        # given:
        converter = _create_mock_string_converter()
        cell_conversion = LinkingDetailCellConversion('product.item.name', converter)

        # and:
        metadata = MetadataEntity(linking_details={
            'item': {'description': 'a jar of milk'}
        })

        # when:
        cell_conversion.apply(metadata, 'milk')

        # then:
        self.assertEqual('milk - converted', metadata.get_linking_detail('item.name'))
        self.assertEqual('a jar of milk', metadata.get_linking_detail('item.description'))



#!/usr/bin/env python
"""
Description goes here
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "25/05/2018"

from unittest import TestCase
from ingest.template.spreadsheet_builder import SpreadsheetBuilder
import unittest
import tests.template.schema_mock_utils as schema_mock
import xlsxwriter
from openpyxl import load_workbook as Reader


class TestSchemaTemplate(TestCase):
    def setUp(self):
        self.longMessage = True
        self.dummyProjectUri = "https://schema.humancellatlas.org/type/project/5.1.0/project"
        self.dummyDonorUri = "https://schema.humancellatlas.org/type/biomaterial/5.1.0/donor_organism"
        pass


    def test_no_schemas(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar", "description" : "this is a foo bar", "example" : "e.g. foo"}} }'

        file = "foo.xlsx"
        spreadsheet_builder = SpreadsheetBuilder(file)
        template = schema_mock.get_template_for_json(data=data)
        spreadsheet_builder._build(template)
        spreadsheet_builder.save_workbook()

        reader = Reader("foo.xlsx")
        sheet = reader["Donor organism"]

        self.assertEqual("this is a foo bar", sheet.cell(row=1, column=1).value)
        self.assertEqual("Foo bar", sheet.cell(row=2, column=1).value)
        self.assertEqual("e.g. foo", sheet.cell(row=3, column=1).value)
        self.assertEqual("donor_organism.foo_bar", sheet.cell(row=4, column=1).value)



    # TODO fixme
    @unittest.skip
    def test_with_tabs_template(self):
        pass
        # spreadsheet_builder.generate_spreadsheet("human_10x.xlsx", tabs_template="tabs_human_10x.yaml",
        #                                          schema_urls=schemas)

    # TODO fixme
    @unittest.skip
    def test_add_columns(self):
        # spreadsheet_builder.generate_spreadsheet("generic.xlsx", schema_urls=schemas)
        pass


    # TODO fixme
    @unittest.skip
    def test_add_sheets(self):
        # spreadsheet_builder.generate_spreadsheet("generic.xlsx", schema_urls=schemas)
        pass



import re

SPLIT_FIELD_REGEX = '(?P<parent>\w*(\.\w*)*)\.(?P<target>\w*)'


def split_field_chain(field):
    parent_path = ''
    target_field = field

    match = re.search(SPLIT_FIELD_REGEX, field)
    if match:
        parent_path = match.group('parent')
        target_field = match.group('target')

    return parent_path, target_field


def extract_root_field(field_chain):
    root_field = None
    if field_chain is not None:
        split = field_chain.split('.')
        root_field = split[0]
    return root_field



#!/usr/bin/env python
"""
Description goes here
"""
import datetime
import hca
import json
import logging
import os
import time
from ingest.utils.s2s_token_client import S2STokenClient
from ingest.utils.token_manager import TokenManager

__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "12/09/2017"

AUTH_INFO_ENV_VAR = "EXPORTER_AUTH_INFO"


class DssApi:
    def __init__(self, url=None):
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.INFO)
        self.logger = logging.getLogger(__name__)

        self.url = url if url else "https://dss.dev.data.humancellatlas.org"
        if not url and 'DSS_API' in os.environ:
            url = os.environ['DSS_API']
            # expand interpolated env vars
            self.url = os.path.expandvars(url)
            self.logger.info("using " + url + " for dss API")

        self.headers = {'Content-type': 'application/json'}

        self.hca_client = hca.dss.DSSClient(swagger_url=f'{self.url}/v1/swagger.json')
        self.hca_client.host = self.url + "/v1"
        self.creator_uid = 8008

    def put_file(self, bundle_uuid, file):
        url = file["url"]
        uuid = file["dss_uuid"]

        version = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S.%fZ")

        update_date = file.get("update_date")

        if update_date:
            update_date = datetime.datetime.strptime(update_date, "%Y-%m-%dT%H:%M:%S.%fZ")
            update_date = update_date.strftime("%Y-%m-%dT%H%M%S.%fZ")
            version = update_date

        # retrying file creation 20 times
        max_retries = 20
        tries = 0
        file_create_complete = False

        params = {
            'uuid': uuid,
            'version': version,
            'bundle_uuid': bundle_uuid,
            'creator_uid': self.creator_uid,
            'source_url': url
        }

        while not file_create_complete and tries < max_retries:
            try:
                tries += 1
                self.logger.info(f'Creating file {file["name"]} in DSS {uuid}:{version} with params: {json.dumps(params)}')
                bundle_file = self.hca_client.put_file(
                    uuid=uuid,
                    version=version,
                    bundle_uuid=bundle_uuid,
                    creator_uid=self.creator_uid,
                    source_url=url
                )
                self.logger.info('Created!')
                file_create_complete = True
                return bundle_file
            except Exception as e:
                self.logger.error(
                    'Attempt {0} out of {1}: Error in hca_client.put_file method call with params:{2} due to {3}'.format(
                        str(tries),
                        str(max_retries),
                        json.dumps(params),
                        str(e))
                )

                if not tries < max_retries:
                    raise Error(e)
                else:
                    time.sleep(60)

    def put_bundle(self, bundle_uuid, bundle_files):
        bundle = None

        # Generate version client-side for idempotent PUT /bundle
        version = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S.%fZ")

        # retrying file creation 20 times
        max_retries = 20
        tries = 0
        bundle_create_complete = False

        # finally create the bundle
        while not bundle_create_complete and tries < max_retries:
            try:
                tries += 1
                self.logger.info(f'Creating bundle in DSS {bundle_uuid}:{version}')
                bundle = self.hca_client.put_bundle(
                    uuid=bundle_uuid,
                    version=version,
                    replica="aws",
                    files=bundle_files,
                    creator_uid=self.creator_uid
                )
                self.logger.info('Created!')
                bundle_create_complete = True
                return bundle
            except Exception as e:
                params = {
                    'uuid': bundle_uuid,
                    'version': version,
                    'replica': "aws",
                    'files': bundle_files,
                    'creator_uid': self.creator_uid
                }
                self.logger.error(
                    'Attempt {0} out of {1}: Error in hca_client.put_bundle method call with params:{2} due to {3}'.format(
                        str(tries),
                        str(max_retries),
                        json.dumps(params),
                        str(e))
                )
                if not tries < max_retries:
                    raise Error(e)
                else:
                    time.sleep(60)

    def head_file(self, file_uuid, version=None):
        # finally create the bundle
        try:
            return self.hca_client.head_file(
                uuid=file_uuid,
                version=version,
                replica="aws"
            )
        except Exception as e:
            raise Error(e)


# Module Exceptions

class Error(Exception):
    """Base-class for all exceptions raised by this module."""